# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: process.proto
"""Generated protocol buffer code."""
# Third Party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\rprocess.proto\x12\x0cprocessproto"\xfe\x02\n\x0eProcessRequest\x12\x12\n\ntrainingID\x18\x01 \x01(\t\x12\x18\n\x10\x63ustomTrainingID\x18\x06 \x01(\t\x12\x43\n\x0crequest_dict\x18\x02 \x03(\x0b\x32-.processproto.ProcessRequest.RequestDictEntry\x12\x1a\n\x12training_input_dir\x18\x03 \x01(\t\x12\x1b\n\x13training_output_dir\x18\x04 \x01(\t\x12\x16\n\x0e\x64\x61ta_input_dir\x18\x05 \x03(\t\x12\x41\n\x0b\x64\x61ta_inputs\x18\x07 \x03(\x0b\x32,.processproto.ProcessRequest.DataInputsEntry\x1a\x32\n\x10RequestDictEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\x31\n\x0f\x44\x61taInputsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01"\xbc\x01\n\x0fProcessResponse\x12\x12\n\ntrainingID\x18\x01 \x01(\t\x12\x18\n\x10\x63ustomTrainingID\x18\x03 \x01(\t\x12\x46\n\rresponse_dict\x18\x02 \x03(\x0b\x32/.processproto.ProcessResponse.ResponseDictEntry\x1a\x33\n\x11ResponseDictEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x32O\n\x07Process\x12\x44\n\x03Run\x12\x1c.processproto.ProcessRequest\x1a\x1d.processproto.ProcessResponse"\x00\x42\x64\n!com.ibm.watson.modeltrain.processZ?github.ibm.com/deep-learning-platform/processproto;processprotob\x06proto3'
)


_PROCESSREQUEST = DESCRIPTOR.message_types_by_name["ProcessRequest"]
_PROCESSREQUEST_REQUESTDICTENTRY = _PROCESSREQUEST.nested_types_by_name[
    "RequestDictEntry"
]
_PROCESSREQUEST_DATAINPUTSENTRY = _PROCESSREQUEST.nested_types_by_name[
    "DataInputsEntry"
]
_PROCESSRESPONSE = DESCRIPTOR.message_types_by_name["ProcessResponse"]
_PROCESSRESPONSE_RESPONSEDICTENTRY = _PROCESSRESPONSE.nested_types_by_name[
    "ResponseDictEntry"
]
ProcessRequest = _reflection.GeneratedProtocolMessageType(
    "ProcessRequest",
    (_message.Message,),
    {
        "RequestDictEntry": _reflection.GeneratedProtocolMessageType(
            "RequestDictEntry",
            (_message.Message,),
            {
                "DESCRIPTOR": _PROCESSREQUEST_REQUESTDICTENTRY,
                "__module__": "process_pb2"
                # @@protoc_insertion_point(class_scope:processproto.ProcessRequest.RequestDictEntry)
            },
        ),
        "DataInputsEntry": _reflection.GeneratedProtocolMessageType(
            "DataInputsEntry",
            (_message.Message,),
            {
                "DESCRIPTOR": _PROCESSREQUEST_DATAINPUTSENTRY,
                "__module__": "process_pb2"
                # @@protoc_insertion_point(class_scope:processproto.ProcessRequest.DataInputsEntry)
            },
        ),
        "DESCRIPTOR": _PROCESSREQUEST,
        "__module__": "process_pb2"
        # @@protoc_insertion_point(class_scope:processproto.ProcessRequest)
    },
)
_sym_db.RegisterMessage(ProcessRequest)
_sym_db.RegisterMessage(ProcessRequest.RequestDictEntry)
_sym_db.RegisterMessage(ProcessRequest.DataInputsEntry)

ProcessResponse = _reflection.GeneratedProtocolMessageType(
    "ProcessResponse",
    (_message.Message,),
    {
        "ResponseDictEntry": _reflection.GeneratedProtocolMessageType(
            "ResponseDictEntry",
            (_message.Message,),
            {
                "DESCRIPTOR": _PROCESSRESPONSE_RESPONSEDICTENTRY,
                "__module__": "process_pb2"
                # @@protoc_insertion_point(class_scope:processproto.ProcessResponse.ResponseDictEntry)
            },
        ),
        "DESCRIPTOR": _PROCESSRESPONSE,
        "__module__": "process_pb2"
        # @@protoc_insertion_point(class_scope:processproto.ProcessResponse)
    },
)
_sym_db.RegisterMessage(ProcessResponse)
_sym_db.RegisterMessage(ProcessResponse.ResponseDictEntry)

_PROCESS = DESCRIPTOR.services_by_name["Process"]
if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    DESCRIPTOR._serialized_options = b"\n!com.ibm.watson.modeltrain.processZ?github.ibm.com/deep-learning-platform/processproto;processproto"
    _PROCESSREQUEST_REQUESTDICTENTRY._options = None
    _PROCESSREQUEST_REQUESTDICTENTRY._serialized_options = b"8\001"
    _PROCESSREQUEST_DATAINPUTSENTRY._options = None
    _PROCESSREQUEST_DATAINPUTSENTRY._serialized_options = b"8\001"
    _PROCESSRESPONSE_RESPONSEDICTENTRY._options = None
    _PROCESSRESPONSE_RESPONSEDICTENTRY._serialized_options = b"8\001"
    _PROCESSREQUEST._serialized_start = 32
    _PROCESSREQUEST._serialized_end = 414
    _PROCESSREQUEST_REQUESTDICTENTRY._serialized_start = 313
    _PROCESSREQUEST_REQUESTDICTENTRY._serialized_end = 363
    _PROCESSREQUEST_DATAINPUTSENTRY._serialized_start = 365
    _PROCESSREQUEST_DATAINPUTSENTRY._serialized_end = 414
    _PROCESSRESPONSE._serialized_start = 417
    _PROCESSRESPONSE._serialized_end = 605
    _PROCESSRESPONSE_RESPONSEDICTENTRY._serialized_start = 554
    _PROCESSRESPONSE_RESPONSEDICTENTRY._serialized_end = 605
    _PROCESS._serialized_start = 607
    _PROCESS._serialized_end = 686
# @@protoc_insertion_point(module_scope)
