# Scheme Base (used for all schemes)
# General merged augmentor schemes to be leveraged by extensions of caikit core
# Local
from .always_selection_scheme import AlwaysSelectionScheme
from .base import SchemeBase
from .random_multi_selection_scheme import RandomMultiSelectionScheme
from .random_single_selection_scheme import RandomSingleSelectionScheme
