# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: model-mesh.proto
"""Generated protocol buffer code."""
# Third Party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x10model-mesh.proto\x12\x05mmesh"\x81\x01\n\x14RegisterModelRequest\x12\x0f\n\x07modelId\x18\x01 \x01(\t\x12#\n\tmodelInfo\x18\x02 \x01(\x0b\x32\x10.mmesh.ModelInfo\x12\x0f\n\x07loadNow\x18\x03 \x01(\x08\x12\x0c\n\x04sync\x18\x04 \x01(\x08\x12\x14\n\x0clastUsedTime\x18\x05 \x01(\x04"4\n\tModelInfo\x12\x0c\n\x04type\x18\x01 \x01(\t\x12\x0c\n\x04path\x18\x02 \x01(\t\x12\x0b\n\x03key\x18\x03 \x01(\t"\xe4\x02\n\x0fModelStatusInfo\x12\x32\n\x06status\x18\x01 \x01(\x0e\x32".mmesh.ModelStatusInfo.ModelStatus\x12\x0e\n\x06\x65rrors\x18\x02 \x03(\t\x12<\n\x0emodelCopyInfos\x18\x03 \x03(\x0b\x32$.mmesh.ModelStatusInfo.ModelCopyInfo\x1ag\n\rModelCopyInfo\x12\x10\n\x08location\x18\x01 \x01(\t\x12\x36\n\ncopyStatus\x18\x02 \x01(\x0e\x32".mmesh.ModelStatusInfo.ModelStatus\x12\x0c\n\x04time\x18\x03 \x01(\x04"f\n\x0bModelStatus\x12\r\n\tNOT_FOUND\x10\x00\x12\x0e\n\nNOT_LOADED\x10\x01\x12\x0b\n\x07LOADING\x10\x02\x12\n\n\x06LOADED\x10\x03\x12\x12\n\x0eLOADING_FAILED\x10\x04\x12\x0b\n\x07UNKNOWN\x10\x05")\n\x16UnregisterModelRequest\x12\x0f\n\x07modelId\x18\x01 \x01(\t"\x19\n\x17UnregisterModelResponse"#\n\x10GetStatusRequest\x12\x0f\n\x07modelId\x18\x01 \x01(\t"P\n\x13\x45nsureLoadedRequest\x12\x0f\n\x07modelId\x18\x01 \x01(\t\x12\x14\n\x0clastUsedTime\x18\x02 \x01(\x04\x12\x0c\n\x04sync\x18\x04 \x01(\x08J\x04\x08\x03\x10\x04"\xce\x02\n\x10VModelStatusInfo\x12\x34\n\x06status\x18\x01 \x01(\x0e\x32$.mmesh.VModelStatusInfo.VModelStatus\x12\x15\n\ractiveModelId\x18\x02 \x01(\t\x12\x15\n\rtargetModelId\x18\x03 \x01(\t\x12\x31\n\x11\x61\x63tiveModelStatus\x18\x04 \x01(\x0b\x32\x16.mmesh.ModelStatusInfo\x12\x31\n\x11targetModelStatus\x18\x05 \x01(\x0b\x32\x16.mmesh.ModelStatusInfo\x12\r\n\x05owner\x18\x06 \x01(\t"a\n\x0cVModelStatus\x12\r\n\tNOT_FOUND\x10\x00\x12\x0b\n\x07\x44\x45\x46INED\x10\x01\x12\x11\n\rTRANSITIONING\x10\x02\x12\x15\n\x11TRANSITION_FAILED\x10\x03\x12\x0b\n\x07UNKNOWN\x10\x05"6\n\x13\x44\x65leteVModelRequest\x12\x10\n\x08vModelId\x18\x01 \x01(\t\x12\r\n\x05owner\x18\x02 \x01(\t"\x16\n\x14\x44\x65leteVModelResponse"\xef\x01\n\x10SetVModelRequest\x12\x10\n\x08vModelId\x18\x01 \x01(\t\x12\r\n\x05owner\x18\n \x01(\t\x12\x15\n\rtargetModelId\x18\x02 \x01(\t\x12\x12\n\nupdateOnly\x18\x03 \x01(\x08\x12#\n\tmodelInfo\x18\x04 \x01(\x0b\x32\x10.mmesh.ModelInfo\x12\x1d\n\x15\x61utoDeleteTargetModel\x18\x05 \x01(\x08\x12\x0f\n\x07loadNow\x18\x06 \x01(\x08\x12\r\n\x05\x66orce\x18\x07 \x01(\x08\x12\x0c\n\x04sync\x18\x08 \x01(\x08\x12\x1d\n\x15\x65xpectedTargetModelId\x18\t \x01(\t"9\n\x16GetVModelStatusRequest\x12\x10\n\x08vModelId\x18\x01 \x01(\t\x12\r\n\x05owner\x18\x02 \x01(\t2\x8b\x04\n\tModelMesh\x12\x46\n\rregisterModel\x12\x1b.mmesh.RegisterModelRequest\x1a\x16.mmesh.ModelStatusInfo"\x00\x12R\n\x0funregisterModel\x12\x1d.mmesh.UnregisterModelRequest\x1a\x1e.mmesh.UnregisterModelResponse"\x00\x12\x43\n\x0egetModelStatus\x12\x17.mmesh.GetStatusRequest\x1a\x16.mmesh.ModelStatusInfo"\x00\x12\x44\n\x0c\x65nsureLoaded\x12\x1a.mmesh.EnsureLoadedRequest\x1a\x16.mmesh.ModelStatusInfo"\x00\x12?\n\tsetVModel\x12\x17.mmesh.SetVModelRequest\x1a\x17.mmesh.VModelStatusInfo"\x00\x12I\n\x0c\x64\x65leteVModel\x12\x1a.mmesh.DeleteVModelRequest\x1a\x1b.mmesh.DeleteVModelResponse"\x00\x12K\n\x0fgetVModelStatus\x12\x1d.mmesh.GetVModelStatusRequest\x1a\x17.mmesh.VModelStatusInfo"\x00\x42(\n\x1c\x63om.ibm.watson.modelmesh.apiP\x01Z\x06/mmeshb\x06proto3'
)


_REGISTERMODELREQUEST = DESCRIPTOR.message_types_by_name["RegisterModelRequest"]
_MODELINFO = DESCRIPTOR.message_types_by_name["ModelInfo"]
_MODELSTATUSINFO = DESCRIPTOR.message_types_by_name["ModelStatusInfo"]
_MODELSTATUSINFO_MODELCOPYINFO = _MODELSTATUSINFO.nested_types_by_name["ModelCopyInfo"]
_UNREGISTERMODELREQUEST = DESCRIPTOR.message_types_by_name["UnregisterModelRequest"]
_UNREGISTERMODELRESPONSE = DESCRIPTOR.message_types_by_name["UnregisterModelResponse"]
_GETSTATUSREQUEST = DESCRIPTOR.message_types_by_name["GetStatusRequest"]
_ENSURELOADEDREQUEST = DESCRIPTOR.message_types_by_name["EnsureLoadedRequest"]
_VMODELSTATUSINFO = DESCRIPTOR.message_types_by_name["VModelStatusInfo"]
_DELETEVMODELREQUEST = DESCRIPTOR.message_types_by_name["DeleteVModelRequest"]
_DELETEVMODELRESPONSE = DESCRIPTOR.message_types_by_name["DeleteVModelResponse"]
_SETVMODELREQUEST = DESCRIPTOR.message_types_by_name["SetVModelRequest"]
_GETVMODELSTATUSREQUEST = DESCRIPTOR.message_types_by_name["GetVModelStatusRequest"]
_MODELSTATUSINFO_MODELSTATUS = _MODELSTATUSINFO.enum_types_by_name["ModelStatus"]
_VMODELSTATUSINFO_VMODELSTATUS = _VMODELSTATUSINFO.enum_types_by_name["VModelStatus"]
RegisterModelRequest = _reflection.GeneratedProtocolMessageType(
    "RegisterModelRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _REGISTERMODELREQUEST,
        "__module__": "model_mesh_pb2"
        # @@protoc_insertion_point(class_scope:mmesh.RegisterModelRequest)
    },
)
_sym_db.RegisterMessage(RegisterModelRequest)

ModelInfo = _reflection.GeneratedProtocolMessageType(
    "ModelInfo",
    (_message.Message,),
    {
        "DESCRIPTOR": _MODELINFO,
        "__module__": "model_mesh_pb2"
        # @@protoc_insertion_point(class_scope:mmesh.ModelInfo)
    },
)
_sym_db.RegisterMessage(ModelInfo)

ModelStatusInfo = _reflection.GeneratedProtocolMessageType(
    "ModelStatusInfo",
    (_message.Message,),
    {
        "ModelCopyInfo": _reflection.GeneratedProtocolMessageType(
            "ModelCopyInfo",
            (_message.Message,),
            {
                "DESCRIPTOR": _MODELSTATUSINFO_MODELCOPYINFO,
                "__module__": "model_mesh_pb2"
                # @@protoc_insertion_point(class_scope:mmesh.ModelStatusInfo.ModelCopyInfo)
            },
        ),
        "DESCRIPTOR": _MODELSTATUSINFO,
        "__module__": "model_mesh_pb2"
        # @@protoc_insertion_point(class_scope:mmesh.ModelStatusInfo)
    },
)
_sym_db.RegisterMessage(ModelStatusInfo)
_sym_db.RegisterMessage(ModelStatusInfo.ModelCopyInfo)

UnregisterModelRequest = _reflection.GeneratedProtocolMessageType(
    "UnregisterModelRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _UNREGISTERMODELREQUEST,
        "__module__": "model_mesh_pb2"
        # @@protoc_insertion_point(class_scope:mmesh.UnregisterModelRequest)
    },
)
_sym_db.RegisterMessage(UnregisterModelRequest)

UnregisterModelResponse = _reflection.GeneratedProtocolMessageType(
    "UnregisterModelResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _UNREGISTERMODELRESPONSE,
        "__module__": "model_mesh_pb2"
        # @@protoc_insertion_point(class_scope:mmesh.UnregisterModelResponse)
    },
)
_sym_db.RegisterMessage(UnregisterModelResponse)

GetStatusRequest = _reflection.GeneratedProtocolMessageType(
    "GetStatusRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _GETSTATUSREQUEST,
        "__module__": "model_mesh_pb2"
        # @@protoc_insertion_point(class_scope:mmesh.GetStatusRequest)
    },
)
_sym_db.RegisterMessage(GetStatusRequest)

EnsureLoadedRequest = _reflection.GeneratedProtocolMessageType(
    "EnsureLoadedRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _ENSURELOADEDREQUEST,
        "__module__": "model_mesh_pb2"
        # @@protoc_insertion_point(class_scope:mmesh.EnsureLoadedRequest)
    },
)
_sym_db.RegisterMessage(EnsureLoadedRequest)

VModelStatusInfo = _reflection.GeneratedProtocolMessageType(
    "VModelStatusInfo",
    (_message.Message,),
    {
        "DESCRIPTOR": _VMODELSTATUSINFO,
        "__module__": "model_mesh_pb2"
        # @@protoc_insertion_point(class_scope:mmesh.VModelStatusInfo)
    },
)
_sym_db.RegisterMessage(VModelStatusInfo)

DeleteVModelRequest = _reflection.GeneratedProtocolMessageType(
    "DeleteVModelRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _DELETEVMODELREQUEST,
        "__module__": "model_mesh_pb2"
        # @@protoc_insertion_point(class_scope:mmesh.DeleteVModelRequest)
    },
)
_sym_db.RegisterMessage(DeleteVModelRequest)

DeleteVModelResponse = _reflection.GeneratedProtocolMessageType(
    "DeleteVModelResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _DELETEVMODELRESPONSE,
        "__module__": "model_mesh_pb2"
        # @@protoc_insertion_point(class_scope:mmesh.DeleteVModelResponse)
    },
)
_sym_db.RegisterMessage(DeleteVModelResponse)

SetVModelRequest = _reflection.GeneratedProtocolMessageType(
    "SetVModelRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _SETVMODELREQUEST,
        "__module__": "model_mesh_pb2"
        # @@protoc_insertion_point(class_scope:mmesh.SetVModelRequest)
    },
)
_sym_db.RegisterMessage(SetVModelRequest)

GetVModelStatusRequest = _reflection.GeneratedProtocolMessageType(
    "GetVModelStatusRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _GETVMODELSTATUSREQUEST,
        "__module__": "model_mesh_pb2"
        # @@protoc_insertion_point(class_scope:mmesh.GetVModelStatusRequest)
    },
)
_sym_db.RegisterMessage(GetVModelStatusRequest)

_MODELMESH = DESCRIPTOR.services_by_name["ModelMesh"]
if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    DESCRIPTOR._serialized_options = (
        b"\n\034com.ibm.watson.modelmesh.apiP\001Z\006/mmesh"
    )
    _REGISTERMODELREQUEST._serialized_start = 28
    _REGISTERMODELREQUEST._serialized_end = 157
    _MODELINFO._serialized_start = 159
    _MODELINFO._serialized_end = 211
    _MODELSTATUSINFO._serialized_start = 214
    _MODELSTATUSINFO._serialized_end = 570
    _MODELSTATUSINFO_MODELCOPYINFO._serialized_start = 363
    _MODELSTATUSINFO_MODELCOPYINFO._serialized_end = 466
    _MODELSTATUSINFO_MODELSTATUS._serialized_start = 468
    _MODELSTATUSINFO_MODELSTATUS._serialized_end = 570
    _UNREGISTERMODELREQUEST._serialized_start = 572
    _UNREGISTERMODELREQUEST._serialized_end = 613
    _UNREGISTERMODELRESPONSE._serialized_start = 615
    _UNREGISTERMODELRESPONSE._serialized_end = 640
    _GETSTATUSREQUEST._serialized_start = 642
    _GETSTATUSREQUEST._serialized_end = 677
    _ENSURELOADEDREQUEST._serialized_start = 679
    _ENSURELOADEDREQUEST._serialized_end = 759
    _VMODELSTATUSINFO._serialized_start = 762
    _VMODELSTATUSINFO._serialized_end = 1096
    _VMODELSTATUSINFO_VMODELSTATUS._serialized_start = 999
    _VMODELSTATUSINFO_VMODELSTATUS._serialized_end = 1096
    _DELETEVMODELREQUEST._serialized_start = 1098
    _DELETEVMODELREQUEST._serialized_end = 1152
    _DELETEVMODELRESPONSE._serialized_start = 1154
    _DELETEVMODELRESPONSE._serialized_end = 1176
    _SETVMODELREQUEST._serialized_start = 1179
    _SETVMODELREQUEST._serialized_end = 1418
    _GETVMODELSTATUSREQUEST._serialized_start = 1420
    _GETVMODELSTATUSREQUEST._serialized_end = 1477
    _MODELMESH._serialized_start = 1480
    _MODELMESH._serialized_end = 2003
# @@protoc_insertion_point(module_scope)
